
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <ctype.h>
#include "xstd/h/strstream.h"
#include "xstd/h/iomanip.h"

#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "base/polyLogCats.h"
#include "runtime/Connection.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/HttpDate.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/SharedOpts.h"
#include "runtime/ConnMgr.h"
#include "runtime/ErrorMgr.h"
#include "runtime/LogComment.h"
#include "runtime/httpText.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentMgr.h"
#include "csm/oid2Url.h"
#include "client/CltOpts.h"
#include "client/Client.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/cltReqTypes.h"
#include "client/CltXact.h"


CltXact::CltXact() {
	CltXact::reset();
}

void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theSrvRep = 0;
	theNextHop = NetAddr();
	theRepHdr.reset();
	theConsumedSize = 0;
	theReqSize = -1;
	thePeerState = peerUnknown;
	doRetry = false;

	theLogCat = lgcCltSide;
}

Agent *CltXact::owner() {
	return theOwner;
}

const UniqId &CltXact::reqId() const {
	return theId;
}

void CltXact::enqueue() {
	theEnqueTime = TheClock;
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(theOid.world());
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);
	theSrvRep = TheHostMap->serverRepAt(theOid.target());
	Assert(theSrvRep);

	makeReq(theConn->theWrBuf);
	theConn->theWr.start(this);
	newState(stSpaceWaiting);
}


void CltXact::finish(Error err) {
	if (!err && staleReply())
		err = errStaleHit;

	// cache objects with bodies
	if (!err && theRepHdr.theContSize >= 0) {
		if (PrivCache *cache = theOwner->privCache())
			cache->putOid(theOid, theRepHdr.calcLmt());
	}

	Xaction::finish(err); // XXX: retries lead to xact.level leak!
	theOwner->noteXactDone(this);
}

void CltXact::noteReadReady(int) {
	theConn->read();

	if (theConn->bad()) {
		doRetry = theState == stHdrWaiting && !theConn->theRdBuf.contSize() && theConn->useCnt() > 1;
		finish(errOther);
		return;
	}

	if (theState == stHdrWaiting) {
		if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {

			// dump reply header
			if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
				printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);

			if (const Error err = interpretHeader()) {
				finish(err);
				return;
			}

			consume(theRepHdr.theHdrSize);
			Assert(theState == stBodyWaiting);
			// fall through
		} else
		if (theConn->theRdBuf.full()) { // too big header
			finish(errHugeHdr);
			return;
		} else
		if (theConn->atEof()) { // premature end of headers
			const bool readNothing = !theConn->theRdBuf.contSize();
			// HTTP pconn race condition?
			if ((doRetry = readNothing && theConn->useCnt() > 1))
				finish(errOther);
			else
				finish(readNothing ? errNoHdrClose : errPrematureEoh);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		const Size parsedSize = parseContent();
		consume(parsedSize);
		theConn->theRdBuf.pack();

		if (theConsumedSize == theRepSize) { // got all we need
			if (theRepHdr.redirect())
				redirect();

			if (theOid.foreign() && !theOid.repToRedir())
				finish(errOther); // see errForeignRep elsewhere
			else
			if (theConn->theRdBuf.contSize() && !theConn->reusable())
				finish(errExtraRepData);
			else
				finish(0);
			return;
		}

		// tag without a closing bracket?
		if (theConn->atEof() &&
			theConsumedSize + theConn->theRdBuf.contSize() >= theRepSize) {
			finish(errOpenTag);
			return;
		}
	}

	if (theConn->atEof()) { // premature eof
		finish(errPrematureEof);
		return;
	}
}

void CltXact::noteWriteReady(int) {
	WrBuf &buf = theConn->theWrBuf;
	Assert(buf.contSize() > 0);
	theConn->write();

	if (theConn->bad()) {
		Assert(theReqSize >= 0);
		// HTTP pconn race condition?
		doRetry = buf.contSize() == theReqSize && theConn->useCnt() > 1;
		finish(errOther);
		return;
	}

	if (buf.empty()) {
		theConn->theWr.stop(this);
		theConn->theRd.start(this);
		TheFileScanner->setPriority(theConn->sock().fd(), fsupBestEffort);
		newState(stHdrWaiting);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

void CltXact::makeReq(WrBuf &buf) {
	const char *reqStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	makeEndToEndHdrs(os);
	makeHopByHopHdrs(os);

	os << crlf; // end-of-header

	theReqSize = os.tellp();
	buf.appended(theReqSize);

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(reqStart, buf.space()-reqStart);
}

void CltXact::makeEndToEndHdrs(ostream &os) {
	const NetAddr &visName = TheViservs.nameAt(theOid.viserv());

	/* request line */
	os << rlpGet;
	if (theOwner->proxy()) { 
		// proxied connection, use full URL
		Oid2Url(theOid, os);
	} else {
		// transparent folks get an advantage of shorter request-lines
		Oid2UrlPath(theOid, os);
	}
	os << rlpHttp_10; // end-of-request-line

	/* request-header fields */
	os 
		<< hfAccept
		<< hfpHost << visName.addrA() << ':' << visName.port() << crlf
		;

	if (theOid.ims()) {
		Assert(!theOid.reload());
		static const Time minTime(0,0);
		const PrivCache *cache = theOwner->privCache();
		const Time stamp = cache ? 
			cache->stamp(theOid) : Clock::TheStartTime;
		HttpDatePrint(os << hfpIMS, Max(minTime, stamp)) << crlf;
	}

	if (theOid.reload()) {
		Assert(!theOid.ims());
		os << hfReload;
	}

	/* entity-header fields */

	os << hfpXXact << TheGroupId << ' ' << theId << crlf;

	// send public world info
	const PubWorld &pubWorld = *ThePubWorlds[theOid.viserv()];
	os << hfpXLocWorld << pubWorld.localSlice() << crlf;
	if (const PubWorldSlice *slice = pubWorld.sliceToSync())
		os << hfpXRemWorld << *slice << crlf;

	os << hfpXTarget << theSrvRep->addr() << crlf;

	// report our readiness to change phase if needed
	if (TheOpts.doSyncPhases)
		os << hfpXPhaseSyncPos << TheStatPhaseMgr.phaseSyncPos() << crlf;
}

void CltXact::makeHopByHopHdrs(ostream &os) {
	/* general-header fields */

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << (theOwner->proxy() ? hfConnAlivePxy : hfConnAliveOrg);
}

void CltXact::consume(Size size) {
	theConn->theRdBuf.consumed(size);
	theConsumedSize += size;
}

Error CltXact::interpretHeader() {

	// does reply line make sense?
	if (theRepHdr.theHttpVer < 0 || theRepHdr.theStatus < 0)
		return errHttpRLine;

	if (const Error err = setStatusCode(theRepHdr.theStatus))
		return err;

	// check content-length, set RepSize
	if (theRepHdr.expectBody()) {
		if (theRepHdr.theContSize < 0)
			return errNoCLen;
		theRepSize = theRepHdr.theHdrSize + theRepHdr.theContSize;
	} else {
		if (theRepHdr.theContSize >= 0)
			return errUnexpectedCLen;
		theRepSize = theRepHdr.theHdrSize;
	}

	if (theRepHdr.redirect()) {
		theOid.repToRedir(true);
		// note: 3xx can be a foreign reply
		theOid.foreign(theRepHdr.foreign());
		// do not count repToRedirs as misses, hits, etc.!
	} else
	if (theRepHdr.foreign()) {
		// note: read foreign replies, but
		// do not count foreign replies as misses, hits, etc.!
		theOid.foreign(true);
		if (ReportError(errForeignRep))
			printMsg(theConn->theRdBuf);
	} else {
		// firstHand here means our request reached the server and
		// server's reply reached us
		const bool firstHand = reqId().myMutant(theRepHdr.theXactId);

		theOid.cachable(theRepHdr.isCachable);
		theOid.hit(!firstHand && theRepHdr.theXactId && !theOid.ims());

		if (theOid.hit()) {
			if (!theOid.offeredHit() && !TheCltOpts.ignoreFalseHits)
				return errFalseHit;

			if (!theOid.cachable())
				return errUnchbHit;

			if (theOid.reload())
				return errReloadHit;
		}

		if (firstHand) {
			checkDateSync();
			firstHandSync();
		}

		if (TheCltOpts.printFalseMisses && theOid.offeredHit() && !theOid.hit()) {
			Oid2Url(theOid, cout << "False-Miss: "); cout << endl;
			printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);
		}
	}
	
	// note if the server will close the connection
	theConn->lastUse(!theRepHdr.keepAlive());

	newState(stBodyWaiting);
	return 0;
}

void CltXact::firstHandSync() {
	// update public world info
	if (theRepHdr.theRemWorld)
		updatePubWorld(theRepHdr.theRemWorld);

	// sync oid accounting on first-hand replies
	theSrvRep->syncSrvInfo(theRepHdr);

	// change phase if needed
	if (TheOpts.doSyncPhases && TheStatPhaseMgr->unlockToStop() && 
		theRepHdr.thePhaseSyncPos >= TheStatPhaseMgr.phaseSyncPos())
		TheStatPhaseMgr->unlock();
}

void CltXact::redirect() {
	ObjId dest = theRepHdr.theLocn.oid;
	dest.rediredReq(true);

	if (!validRelOid(dest)) {
		ReportError(errRedirLocation);
		return;
	}

	if (theRepHdr.theLocn.host) {
		if (const Error err = setViserv(theRepHdr.theLocn.host, dest))
			return;
	} else {
		dest.viserv(theOid.viserv());
	}
	
	theOwner->selectTarget(dest);
	dest.repeat(true); // for now, server only redirects to seen URLs
	// XXX: hot() is unset

	theOwner->noteRedirect(this, dest);
}

bool CltXact::validRelOid(const ObjId &oid) const {
	return oid.world() && oid.name() > 0 && 
		TheContentMgr.validId(oid.type());
}

bool CltXact::staleReply() const {
	return theRepHdr.theExpires >= 0 && 
		theRepHdr.theExpires < theRepHdr.theDate;
}

Error CltXact::setStatusCode(int aStatus) {
	theHttpStatus = aStatus;
	if (theHttpStatus != RepHdr::sc200_OK && 
		theHttpStatus != RepHdr::sc304_NotModified &&
		!theRepHdr.redirect())
		return errHttpStatusCode;
	return Error();
}

void CltXact::checkDateSync() {
	if (theRepHdr.theDate < 0) {
		ReportError(errHttpNoDate);
		return;
	}

	const Time maxGap = Time::Sec(60);

	// cannot measure drift using replies that took too long
	if (TheClock - theStartTime >= maxGap/2)
		return;

	if (theRepHdr.theDate > TheClock.time() + maxGap ||
		theRepHdr.theDate < TheClock.time() - maxGap) {
		if (ReportError(errSyncDate)) {
			HttpDatePrint(Comment << "request generated:  ", theStartTime) << endc;
			HttpDatePrint(Comment << "response generated: ", theRepHdr.theDate) << endc;
			HttpDatePrint(Comment << "response received:  ", TheClock.time()) << endc;
			const Time diff = theRepHdr.theDate - TheClock.time();
			Comment << "time difference:  " << diff << endc;
		}
		return;
	}
}

void CltXact::updatePubWorld(const ObjWorld &slice) {
	PubWorld &pubWorld = *ThePubWorlds[theOid.viserv()];
	int sliceIdx = -1;
	if (pubWorld.find(slice.id(), sliceIdx)) {
		pubWorld.sliceAt(sliceIdx).update(slice);
	} else {
		pubWorld.add(slice);
	}
}

// returns number of bytes that can be consumed
Size CltXact::parseContent() {
	const IOBuf &buf = theConn->theRdBuf;
	const char *beg = buf.content();
	const char *end = beg + Min(buf.contSize(), theRepSize - theConsumedSize);
	const char *p = beg;

	// do not attempt to parse foreign content
	// do not attempt to parse embedded content
	if (theOid.foreign() || theOid.embedded())
		return end - beg;

	while (p < end) {
		// find next tag opening
		if (const char *tagO = StrBoundChr(p, '<', end)) {
			p = tagO++;
			// find tag closing
			if (const char *tagC = StrBoundChr(tagO, '>', end)) {
				// found complete tag, parse it
				parseTag(tagO, tagC);
				p = tagC+1;
				// continue
			} else
				break; // wait for more data, leave tag prefix
		} else {
			p = end; // no more tags till the end of a buffer
			break;
		}
	}

	return p - beg;
}

void CltXact::parseTag(const char *tagB, const char *tagE) {
	Error err;

	// find "src=" token within a tag
	for (const char *p = tagB;  p && tagE - p >= 6; p = StrBoundAfterSpace(p, tagE)) {

		if (strncasecmp(p, "src=\"", 5) == 0) {
			const char *url = p + 5;
			if ((p = StrBoundChr(url, '"', tagE))) {
				ReqHdr h;
				if (h.parseUri(url, p, h.theUri) && validRelOid(h.theUri.oid)) {

					if (h.theUri.host)
						err = setViserv(h.theUri.host, h.theUri.oid);
					else
						h.theUri.oid.viserv(theOid.viserv()); // relative URL

					if (!err) {
						// new oid record needs fixing
						theOwner->selectTarget(h.theUri.oid);
						h.theUri.oid.repeat(theOid.repeat());
						h.theUri.oid.hot(theOid.hot());
						h.theUri.oid.ims(theOid.ims());
						h.theUri.oid.reload(theOid.reload());
						theOwner->noteEmbedded(this, h.theUri.oid);
						return;
					}
				} else
					err = errForeignTag;
			} else
				err = errMalformedTag;
			break;
		}
	}

	if (err && !TheCltOpts.ignoreBadContTags &&
		ReportError(err) && TheOpts.theDumpFlags(dumpErr, dumpAny))
		printMsg(tagB, tagE - tagB);
}

Error CltXact::setViserv(const NetAddr &name, ObjId &oid) const {
	// XXX: merge with SrvXact::setViserv() ?
	int nameIdx = -1;
	if (!TheAddrMap->find(name, nameIdx))
		return errForeignHostName;

	int viserv = -1;
	if (!TheViservs.find(nameIdx, viserv))
		return errSrvRedirect; // disallow for now

	if (viserv != theOid.viserv())
		return errSrvRedirect; // disallow for now

	oid.viserv(viserv);
	return Error();
}

void CltXact::logStats(OLog &ol) const {
	Xaction::logStats(ol);
	ol << theOwner->seqvId() << theSrvRep->serverIdx();
}
