
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/h/math.h"
#include <limits.h>

#include "pgl/PglBoolSym.h"
#include "pgl/PglNumSym.h"
#include "pgl/PglTimeSym.h"
#include "pgl/PglSizeSym.h"
#include "pgl/PglRateSym.h"
#include "pgl/PglBwidthSym.h"
#include "pgl/PglDistrSym.h"
#include "pgl/PglArraySym.h"
#include "pgl/PglIntSym.h"

#include "xstd/gadgets.h"

IntSym::IntSym(int aVal = -1): ExpressionSym("int"), theVal(aVal) {
}

bool IntSym::isA(const String &type) const {
	return ExpressionSym::isA(type) ||
		type == "Int" || type == "int";
}

SynSym *IntSym::dupe(const String &type) const {
	if (isA(type))
		return new IntSym(theVal);

	if (type == "float")
		return new NumSym(theVal);

	if (type == "bool")
		return new BoolSym(theVal != 0);

	// zero is zero regardless of "scale"
	if (!theVal) {
		if (type == "time")
			return new TimeSym(Time(0,0));
		if (type == "size")
			return new SizeSym(BigSize(0));
		if (type == "rate")
			return new RateSym(0);
		if (type == "bwidth")
			return new BwidthSym(0);
		if (type.str("_distr"))
			return new DistrSym(type, 0);
	}

	return ExpressionSym::dupe(type);
}

ExpressionSym *IntSym::unOper(const String &op) const {
	if (op == "+")
		return new IntSym(+theVal);
	else
	if (op == "-")
		return new IntSym(-theVal);
	else
	if (op == "!")
		return new BoolSym(!theVal);
	else
		return ExpressionSym::unOper(op);
}


ExpressionSym *IntSym::bnOper(const String &op, const SynSym &s) const {
	double otherVal = 0;

	if (s.isA("int")) {
		otherVal = ((const IntSym&)s.cast("int")).val();
	} else 
	if (IntSym *is = (IntSym*)s.clone("int")) {
		otherVal = is->val();
		delete is;
	} else
		return ExpressionSym::bnOper(op, s);

	if (op.len() != 1)
		return op == ".." ? operRange((int)otherVal) :
			ExpressionSym::bnOper(op, s);

	if (op[0] == '/' && !(int)otherVal)
		cerr << theLoc << "denominator is zero" << endl << xexit;

	switch (op[0]) {
		case '+': return fit(op, theVal + otherVal);
		case '-': return fit(op, theVal - otherVal);
		case '*': return fit(op, theVal * otherVal);
		case '/': return fit(op, theVal / otherVal);
		case '^': return fit(op, pow(theVal, otherVal));
		default: return ExpressionSym::bnOper(op, s);
	}
}

IntSym *IntSym::fit(const String &op, double d) const {
	return Fit(op, d, theLoc);
}

ExpressionSym *IntSym::operRange(int otherEnd) const {
	ArraySym *arr = new ArraySym("int");

	const int step = theVal <= otherEnd ? +1 : -1;
	for (int i = theVal; step*i <= step*otherEnd; i += step) {
		IntSym s(i);
		arr->add(s);
	}

	Assert(arr->count()); // empty ranges are probably impossible
	return arr;
}

ostream &IntSym::print(ostream &os, const String &) const {
	return os << theVal;
}


IntSym *IntSym::Fit(const String &op, double d, const TokenLoc &loc) {
	if (fabs(d) > INT_MAX)
		cerr << loc << "integer overflow as a result of `" 
			<< op << "' (|" << d << "| > " << INT_MAX << ")" 
			<< endl << xexit;
	return new IntSym((int)d);
}
