
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/h/strstream.h"

#include "xstd/String.h"
#include "xstd/NetAddr.h"
#include "xstd/gadgets.h"
#include "pgl/PglNetAddrArraySym.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PglNetAddrLexer.h"
#include "pgl/PglNetAddrBlock.h"
#include "pgl/PglNetAddrRange.h"
#include "pgl/pglStrIs.h"


PglNetAddrRange::PglNetAddrRange(): theSubnet(-1) {
}

PglNetAddrRange::~PglNetAddrRange() {
	while(theBlocks.count()) delete theBlocks.pop();
}

void PglNetAddrRange::reset() {
	theBlocks.reset();
	theAddr = 0;
	theIfName = 0;
	theSubnet = -1;
}

int PglNetAddrRange::addrCount() const {
	int count = 0;
	startIter();
	do {
		count++;
	} while (nextIter());
	return count;
}

void PglNetAddrRange::netmask(struct in_addr &netm) const {
	int subnet = theSubnet; // explicitly specified subnet, if any

	if (subnet < 0)
		subnet = rangeCnt() > 1 ? 16 : 24;

	netm = NetAddr::NetMask(subnet);
}

String PglNetAddrRange::toStr() const {
	ostrstream buf;

	if (theIfName)
		buf << theIfName << "::";

	for (int i = 0; i < theBlocks.count(); ++i)
		theBlocks[i]->print(buf);

	if (theSubnet >= 0)
		buf << '/' << theSubnet;

	buf << ends;

	const String res = buf.str();
	ostrstreamFreeze(buf, false);
	return res;
}

void PglNetAddrRange::toStrs(Array<String*> &strs) const {
	startIter();
	do {
		String str;
		currentIterAddr(str);
		strs.append(new String(str));
	} while (nextIter());
}

ArraySym *PglNetAddrRange::toSyms() const {
	ArraySym *arr = new NetAddrArraySym();

	startIter();
	do {
		String str;
		currentIterAddr(str);

		NetAddr addr = pglIsNetAddr(str);
		Assert(addr);

		NetAddrSym nas;
		nas.val(addr);
		if (theIfName)
			nas.setIfname(theIfName);
		if (theSubnet >= 0)
			nas.setSubnet(theSubnet);
		arr->add(nas);

	} while (nextIter());

	return arr;
}

void PglNetAddrRange::toAddrs(Array<NetAddr*> &addrs) const {
	startIter();
	do {
		String str;
		currentIterAddr(str);

		NetAddr addr = pglIsNetAddr(str);
		Assert(addr);

		addrs.append(new NetAddr(addr));
	} while (nextIter());
}

void PglNetAddrRange::toAddrs(AddrIter iter) const {
	startIter();
	do {
		String str;
		currentIterAddr(str);

		NetAddr addr = pglIsNetAddr(str);
		Assert(addr);

		iter(addr);
	} while (nextIter());
}

void PglNetAddrRange::startIter() const {
	for (int i = 0; i < theBlocks.count(); ++i)
		theBlocks[i]->start();
}

bool PglNetAddrRange::nextIter() const {
	return nextIter(theBlocks.count()-1);
}

bool PglNetAddrRange::nextIter(int level) const {
	if (level < 0)
		return false;

	PglNetAddrBlock &b = *theBlocks[level];
	// delegate to next level if overflow
	if (b.atLast()) {
		b.start();
		return nextIter(level-1);
	}

	b.next();
	return true;
}

void PglNetAddrRange::currentIterAddr(String &str) const {
	ostrstream buf;
	for (int i = 0; i < theBlocks.count(); ++i) {
		theBlocks[i]->printCur(buf);
	}
	buf << ends;
	str = buf.str();
	ostrstreamFreeze(buf, false);
}

bool PglNetAddrRange::parse(const String &val) {
	if (!pglSplitAddr(val, theIfName, theAddr, theSubnet))
		return false;

	PglNetAddrLexer lexs(this, theAddr);

	// convert val into lexems
	if (!lexs.parse())
		return false;

	// convert lexems into blocks
	while (lexs) {
		if (lexs.next() == '.') {
			lexs.step();
			(void)lexs.range();
		} else
		if (lexs.next() == '[') {
			lexs.skip();
			if (lexs.range() && lexs.next() == ']')
				lexs.skip();
			else
				return false; // invalid range specs
		} else
		if (lexs.next() == ':') { // port(s)
			lexs.step();
			(void)lexs.range();
		} else
		if (lexs.next() == ']') {
			return false; // unbalanced ']'
		} else {
			lexs.step(); // simple label
		}
	}

	return true;
}

void PglNetAddrRange::addStrBlock(const char *start, const char *stop) {
	theBlocks.append(new PglStrAddrBlock(start, stop));
}

void PglNetAddrRange::addRangeBlock(int start, int stop) {
	theBlocks.append(new PglRangeAddrBlock(start, stop));
}

bool PglNetAddrRange::canMerge(const PglNetAddrRange &r) const {
	if (!theAddr) // brand new object
		return true;

	if (theBlocks.count() != r.theBlocks.count())
		return false;

	if ((theIfName || r.theIfName) && theIfName != r.theIfName)
		return false;

	if (theSubnet != r.theSubnet)
		return false;

	int diffCount = 0;
	for (int i = 0; diffCount <= 1 && i < theBlocks.count(); ++i) {
		const PglNetAddrBlock &b1 = *theBlocks[i];
		const PglNetAddrBlock &b2 = *r.theBlocks[i];
		diffCount += b1.diffCount(b2);
	}

	// can merge ranges that differ by one block only because 
	// the range notation cannot express other resulting sets
	return diffCount == 1;
}

void PglNetAddrRange::merge(const PglNetAddrRange &r) {
	if (!theAddr) { // brand new object
		//theAddr = r.theAddr;
		theIfName = r.theIfName;
		theSubnet = r.theSubnet;
		theBlocks.stretch(r.theBlocks.count());
		for (int i = 0; i < r.theBlocks.count(); ++i)
			theBlocks.append(r.theBlocks[i]->clone());
	} else {
		Assert(theBlocks.count() == r.theBlocks.count());
		for (int i = 0; i < theBlocks.count(); ++i)
			theBlocks[i]->merge(*r.theBlocks[i]);
		//theAddr = toStr();
	}

	// we must rebuild the range because when string blocks
	// are merged the start/stop pointers may point to memory
	// in the callers stack (i.e. "r" may be on the stack)
	const String s = toStr();
	reset();
	Assert(parse(s));
}

int PglNetAddrRange::rangeCnt() const {
	int count = 0;
	for (int i = 0; i < theBlocks.count(); ++i) {
		if (theBlocks[i]->count() > 1)
			count++;
	}
	return count;
}
