
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <limits.h>
#include "xstd/h/strstream.h"

#include "xstd/Clock.h"
#include "xstd/Rnd.h"
#include "xstd/gadgets.h"
#include "base/RndPermut.h"
#include "base/ObjId.h"
#include "base/polyLogCats.h"
#include "runtime/Connection.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/HttpDate.h"
#include "runtime/ErrorMgr.h"
#include "runtime/SharedOpts.h"
#include "runtime/PopModel.h"
#include "runtime/LogComment.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentMgr.h"
#include "csm/ContentCfg.h"
#include "csm/BodyIter.h"
#include "csm/ObjLifeCycle.h"
#include "csm/oid2Url.h"
#include "server/CltGrpGlobMap.h"
#include "server/Server.h"
#include "server/SrvCfg.h"
#include "server/SrvXact.h"
#include "server/SrvOpts.h"

/* HTTP header strings (to avoid strlen() calls when creating headers) */

// status lines
static const String slHttp_10_200 = "HTTP/1.0 200 OK\r\n";       
static const String slHttp_10_302 = "HTTP/1.0 302 Found\r\n";
static const String slHttp_10_304 = "HTTP/1.0 304 Not Modified\r\n";

// full header fields
static const String hfCcCachable  = "Cache-Control: public\r\n";
static const String hfCcUncachable= "Cache-Control: private,no-cache\r\n"
                                    "Pragma: no-cache\r\n";
static const String hfConnAlive   = "Connection: keep-alive\r\n";
static const String hfConnClose   = "Connection: close\r\n";

// header field prefixes
static const String hfpContLength = "Content-Length: ";
static const String hfpContType   = "Content-Type: ";
static const String hfpDate       = "Date: ";
static const String hfpExpires    = "Expires: ";
static const String hfpLmt        = "Last-Modified: ";
static const String hfpLocation   = "Location: ";
static const String hfpXRemWorld  = "X-Rem-World: ";
static const String hfpXXact      = "X-Xact: ";
static const String hfpXPhaseSyncPos = "X-Phase-Sync-Pos: ";
static const String hfpXTarget    = "X-Target: ";

// free text
static const String text302Found  = "please go to ";


void SrvXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theContentCfg = 0;
	theBodyIter = 0;
	theReqHdr.reset();
	theTimes.reset();
	theRepHdrSize = -1;
	theProducedSize = 0;
	theSeed = -1;

	theLogCat = lgcSrvSide;
}

Agent *SrvXact::owner() {
	return theOwner;
}

void SrvXact::exec(Server *anOwner, Connection *aConn, Time delay) {
	Assert(anOwner);
	theOwner = anOwner;
	theConn = aConn;
	if (delay > 0)
		sleepFor(delay);
	else
		doStart();
}

void SrvXact::doStart() {
	start(theConn);

	newState(stHdrWaiting);

	Assert(!theConn->theRd.theReserv); // cleared in conn mgr
	// note: we assume that conn mgr read() on first noteReadReady()
	noteDataReady(); // push state machine forward
}

void SrvXact::finish(Error err) {
	Xaction::finish(err);
	if (theBodyIter)
		theContentCfg->putBodyIter(theBodyIter);
	theOwner->noteXactDone(this);
}

void SrvXact::wakeUp(const Alarm &a) {
	AlarmUser::wakeUp(a);
	doStart();
}

void SrvXact::noteReadReady(int) {
	Assert(theState == stHdrWaiting);

	theConn->read();

	if (theConn->bad())
		finish(errOther);
	else
		noteDataReady();
}

void SrvXact::noteDataReady() {
	if (theReqHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {
		if (theConn->theRd.theReserv)
			theConn->theRd.stop(this);
		interpretHeader();
		return;
	}

	if (theConn->theRdBuf.full()) { // too big header
		finish(errHugeHdr);
		return;
	}

	if (theConn->atEof()) {         // premature end of headers
		finish(errPrematureEoh);
		return;
	}

	if (!theConn->theRd.theReserv)  // need to read more
		theConn->theRd.start(this);
}

void SrvXact::interpretHeader() {
	// XXX: pxy server cannot support ignoreUrls; see Server::hostIdx
	const bool ignoreUrls = theReqHdr.isHealthCheck ||
		(TheSrvOpts.ignoreUrls && theOwner->hostIdx() >= 0);

	if (ignoreUrls) {
		// XXX: theOid.visName() will not be set!
		// we have to ignore host settings;
		// should we create our own worlds??
		static RndGen rng;
		theReqHdr.theUri.oid.world().create();
		theReqHdr.theUri.oid.target(theOwner->hostIdx());
		theReqHdr.theUri.oid.type(1);
		theReqHdr.theUri.oid.name(rng.ltrial() | 1);
		theReqHdr.theUri.oid.foreign(false);

		// XXX: report healthchecks once in a while and collect stats
		if (theReqHdr.isHealthCheck) {
			static bool didOnce = false;
			if (!didOnce) {
				Comment(6) << "fyi: first health check received from " << theConn->raddr() << ':' << endc;
				printMsg(theConn->theRdBuf, theReqHdr.theHdrSize);
				didOnce = true;
			}
		}
	}

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(theConn->theRdBuf, theReqHdr.theHdrSize);

	if (theReqHdr.theUri.oid.foreign()) {
		theOid.foreign(true);
		finish(errForeignReq); // foreign request
		return;
	}

	// note: clt and srv code assume we consume only after parse!
	theConn->theRdBuf.consumed(theReqHdr.theHdrSize);
	theConn->theRdBuf.pack();

	theOid = theReqHdr.theUri.oid;

	if (!ignoreUrls) {
		if (!theReqHdr.theUri.host) {
			finish(errNoHostName);
			return;
		}

		if (!setViserv(theReqHdr.theUri.host))
			return; // setViserv called finish

		if (!theReqHdr.theTarget) {
			finish(errNoTarget);
			return;
		}

		if (!setTarget(theReqHdr.theTarget))
			return; // setTarget called finish
		Assert(theOid.target() >= 0);

		if (!checkUri())
			return; // checkUri called finish
	}

	// update history for this server
	if (theOid.viserv() >= 0 && theReqHdr.theLocWorld)
		updatePubWorld(theReqHdr.theLocWorld);

	// update client group info
	if (theReqHdr.thePhaseSyncPos >= 0 && theReqHdr.theGroupId)
		TheCltGrpGlobMap.notePhaseSync(theReqHdr.theGroupId, theReqHdr.thePhaseSyncPos);

	// note if the client will close the connection
	theConn->lastUse(!theReqHdr.keepAlive());

	theOid.ims(theReqHdr.theIms >= 0);
	theOid.reload(!theReqHdr.isCachable);

	// get content specs for this reply
	Assert(theOid.type() > 0);
	theContentCfg = TheContentMgr.get(theOid.type());

	theRepHdrSize = -1; // so that we know that we need to make a reply

	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);

	newState(stSpaceWaiting);
}

void SrvXact::noteWriteReady(int) {
	Assert(theState == stSpaceWaiting);
	WrBuf &buf = theConn->theWrBuf;

	if (theRepHdrSize < 0)
		makeRep(buf);

	Assert(theProducedSize > 0 || buf.contSize() > 0);
	
	// keep the write buffer full
	if (theBodyIter)
		theBodyIter->pour();
	else
		Assert(theProducedSize + buf.contSize() >= theRepSize);

	//cerr << here << "noteWriteReady: " << theProducedSize << " < " << theRepSize << endl;

	const Size sz = theConn->write();
	if (theConn->bad()) {
		finish(errOther);
		return;
	}

	Assert(sz >= 0);
	theProducedSize += sz;
	if (theProducedSize >= theRepSize) { // done
		Assert(theProducedSize == theRepSize);
		finish(0);
		return;
	}

	if (!theConn->theWr.theReserv) // need write more
		theConn->theWr.start(this);
}

// decide what kind of reply to build
void SrvXact::makeRep(WrBuf &buf) {
	Assert(theContentCfg);

	// timestamps and cachability are used for all types of replies
	theSeed = GlbPermut(theOid.hash(), rndSrvSeed);
	theContentCfg->theObjLifeCycle->calcTimes(theSeed, theTimes);
	theOid.cachable(calcCachability());

	theOwner->selectRepType(theOid);

	const char *repStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	if (shouldMake302Found() && make302Found(os))
		; // nothing to be done here
	else
	if (shouldMake304NotMod())
		make304NotMod(os);
	else
		make200OK(os);

	buf.appended(os.tellp());

	Assert(theRepHdrSize > 0 && theRepSize >= theRepHdrSize);

	// dump reply header
	if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
		printMsg(repStart, theRepHdrSize);
}

bool SrvXact::shouldMake302Found() const {
	return theOid.repToRedir();
}

bool SrvXact::canMake302Found(ObjId &oid) const {
	Assert(theOwner->popModel());

	// at this time redirect to the same server only
	if (theOid.viserv() < 0)
		return false;

	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	if (!world.canRepeat())
		return false;

	static RndGen rng;
	oid = theOid; // set viserv and such, name/type will be overwritten
	world.repeat(oid, theOwner->popModel(), rng);
	oid.type(Oid2ContType(theOid));
	return true;
}

bool SrvXact::make302Found(ostream &os) {
	ObjId newOid;

	// make sure we have an object to redirect to
	if (!canMake302Found(newOid)) {
		ReportError(errMake302Found);
		return false;
	}

	theHttpStatus = RepHdr::sc302_Found;

	os << slHttp_10_302;
	putStdFields(os);

	os << hfpLocation;
	const Size urlStart = os.tellp();
	Oid2Url(newOid, os);
	const Size urlLen = Size(os.tellp()) - urlStart;
	os << crlf;

	const Size clen = urlLen + Size(text302Found.len());
	os << hfpContLength << (int)clen << crlf;

	os << crlf; // end-of-headers

	theRepHdrSize = os.tellp();

	// text for humans
	Oid2Url(newOid, os << text302Found);

	theRepSize = theRepHdrSize + clen;
	return true;
}

bool SrvXact::shouldMake304NotMod() const {
	return 
		theReqHdr.theIms >= 0 && theTimes.lmt() >= 0 && 
		theTimes.lmt() <= theReqHdr.theIms;
}

void SrvXact::make304NotMod(ostream &os) {
	theHttpStatus = RepHdr::sc304_NotModified;

	os << slHttp_10_304;
	putStdFields(os);
	os << crlf; // end-of-headers

	theRepHdrSize = theRepSize = os.tellp();
}

// build headers for a happy "200 OK" reply
void SrvXact::make200OK(ostream &os) {
	theHttpStatus = RepHdr::sc200_OK;

	os << slHttp_10_200;
	const Size clen = put200OkBody(os);
	os << crlf; // end-of-headers

	theRepHdrSize = os.tellp();
	theRepSize = theRepHdrSize + clen;
	theBodyIter = theContentCfg->getBodyIter(theOid);
	theBodyIter->configure(&theConn->theWrBuf, clen);
}

Size SrvXact::put200OkBody(ostream &os) {
	/* it is "good practice" to send general-header fields first,
	 * followed by request-header or response-header fields, and
	 * ending with the entity-header fields. */

	/* general-header fields */

	os << (theOid.cachable() ? hfCcCachable : hfCcUncachable);

	putStdFields(os);

	/* other entity-header fields */

	if (theTimes.showLmt())
		HttpDatePrint(os << hfpLmt, theTimes.lmt()) << crlf;

	const Size clen = calcRepSize();
	os << hfpContLength << (int)clen << crlf;

	if (theContentCfg->theMimeType)
		os << hfpContType << theContentCfg->theMimeType << crlf;

	/* extention-header fields (they are entity-header fields as well) */

	// put group ids with source/target ids on one line?
	os << hfpXTarget << theOwner->host() << crlf;

	if (theReqHdr.theXactId)
		os << hfpXXact 
			<< TheGroupId
			<< ' ' << theReqHdr.theXactId.genMutant()
			<< crlf;

	if (theOid.viserv() >= 0 && theReqHdr.theRemWorld)
		putRemWorld(os, theReqHdr.theRemWorld);

	os << hfpXPhaseSyncPos << TheCltGrpGlobMap.phaseSyncPos() << crlf;

	return clen;
}

// put fields acceptible for both 304 and 200 replies
void SrvXact::putStdFields(ostream &os) const {
	// general-header fields

	HttpDatePrint(os << hfpDate) << crlf;

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << hfConnAlive;

	// response-header fields (none)
	// entity-header fields

	if (theTimes.knownExp())
		HttpDatePrint(os << hfpExpires, theTimes.exp()) << crlf;
}

Size SrvXact::calcRepSize() const {
	// seed random generator and generate random size
	RndDistr &repSzD = *theContentCfg->theSize;
	repSzD.rndGen()->seed(theSeed);
	const double dh = repSzD();
	// prevent int overflows and leave room for the HTTP header
	const Size sz = (int)Min(ceil(dh), (double)INT_MAX - 100*1024);
	Assert(sz >= 0);
	return sz;
}

bool SrvXact::calcCachability() const {
	RndGen rng(theSeed);
	return rng.event(theContentCfg->theChbRatio);
}


bool SrvXact::setViserv(const NetAddr &name) {
	// XXX: the lookups below are very slow
	int nameIdx = -1;
	if (!TheAddrMap->find(name, nameIdx)) {
		finish(errForeignHostName);
		return false;
	}

	int viserv = -1;
	if (!TheViservs.find(nameIdx, viserv)) {
		TheViservs.add(nameIdx);
		ThePubWorlds.append(new PubWorld());
	}

	theOid.viserv(viserv);
	return true;
}

bool SrvXact::setTarget(const NetAddr &) {
	// XXX: the check below should know about DNS RR, etc.
	//if (target != theOwner->host())
	//	ReportError(errMisdirRequest);

	theOid.target(theOwner->hostIdx());
	return true;
}

// check that requested URI belongs to our server
bool SrvXact::checkUri() {
	const int pathLen = theReqHdr.theUri.pathLen;
	Assert(pathLen >= 0);
	Assert(theReqHdr.theUri.pathBuf);

	// first check if the content type belongs to us
	if (!theOwner->cfg()->hasContType(theOid.type())) {
		if (ReportError(errMisdirRequest)) {
			Comment << "host " << theOwner->host()
				<< " does not have content type implied by request URI: ";
			Comment.write(theReqHdr.theUri.pathBuf, pathLen);
			Comment << endc;
		}
		finish(errOther);
		return false;
	}

	ostrstream os; // XXX: expensive mallocs
	Oid2UrlPath(theOid, os);
	os << ends;

	const char *buf = os.str();
	const bool res = (pathLen+1 == (int)os.tellp()) &&
		strncmp(buf, theReqHdr.theUri.pathBuf, pathLen) == 0;
	ostrstreamFreeze(os, false);

	if (!res) {
		if (ReportError(errMisdirRequest)) {
			Comment << "host: " << theOwner->host() << endc;
			(Comment << "received: ").write(theReqHdr.theUri.pathBuf, pathLen);
			Comment << endc;
			Oid2UrlPath(theOid, Comment << "expected: ");
			Comment << endc;
		}
		finish(errOther);
	}
	return res;
}

void SrvXact::updatePubWorld(const ObjWorld &newSlice) {
	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	int sliceIdx;
	if (!world.find(newSlice.id(), sliceIdx)) {
		world.add(newSlice);
	} else {
		world.sliceAt(sliceIdx).update(newSlice);
	}
}

void SrvXact::putRemWorld(ostream &os, const ObjWorld &oldSlice) {
	PubWorld &world = *ThePubWorlds[theOid.viserv()];

	int sliceIdx;
	if (world.find(oldSlice.id(), sliceIdx)) {
		if (const PubWorldSlice *slice = world.newerSlice(oldSlice, sliceIdx))
			os << hfpXRemWorld << *slice << crlf;
	} 

	// else client-side sent remote world ID before the same public world ID
}

void SrvXact::logStats(OLog &ol) const {
	Xaction::logStats(ol);
	ol << theOwner->seqvId();
	theTimes.store(ol);
	ol << (int)theReqHdr.theIms.sec() << theSeed;
}
